import requests

from cached_property import cached_property
from google.auth.transport.requests import AuthorizedSession
from ostree.remote import RemoteAuth, RemoteImage
from google.oauth2.service_account import Credentials


class DockerIdentity(object):

    __slots__ = ('name', 'host', 'repository', 'tag', 'digest')

    def __init__(self, host, repository, tag, digest):
        self.host = host
        self.repository = repository
        self.tag = tag
        self.digest = digest

    @classmethod
    def from_name(cls, name):
        # See https://stackoverflow.com/q/37861791
        if '@' in name:
            name, digest = name.split('@')
        else:
            digest = None

        parts = name.split('/')

        if '/' in name and ('.' in parts[0] or ':' in parts[0]):
            host, parts = parts[0], parts[1:]
        else:
            host = None

        if ':' in parts[-1]:
            parts[-1], tag = parts[-1].split(':')
        else:
            tag = 'latest'

        repository = '/'.join(parts)

        return cls(host, repository, tag, digest)


class PublicDockerAuth(RemoteAuth, hosts=(None, )):

    def create_session(self, auth):
        token = auth or self.acquire_authorization_token()

        session = requests.Session()
        session.headers['Authorization'] = f'Bearer {token}'

        return session

    def acquire_authorization_token(self):
        return requests.get(
            f'https://auth.docker.io/token'
            f'?service=registry.docker.io'
            f'&scope=repository:library/{self.identity.repository}:pull'
        ).json()['token']


class GCRAuth(RemoteAuth, hosts=(
    'gcr.io',
    'asia.gcr.io',
    'eu.gcr.io',
    'us.gcr.io',
)):

    def create_session(self, auth):
        if not auth:
            return requests.Session()

        credentials = Credentials.from_service_account_file(auth).with_scopes(
            ['https://www.googleapis.com/auth/cloud-platform']
        )

        return AuthorizedSession(credentials)


class RemoteDockerImage(RemoteImage, protocol='docker'):

    def __init__(self, name, auth=None):
        self.identity = DockerIdentity.from_name(name)
        self.auth = auth

    @property
    def accept_header(self):
        # ostree currently limits itself to this v2 docker manifests
        return 'application/vnd.docker.distribution.manifest.v2+json'

    @cached_property
    def auth_service(self):
        return RemoteAuth.from_identity(self.identity)

    @cached_property
    def registry_host(self):
        return self.identity.host or 'registry-1.docker.io'

    @cached_property
    def reference(self):
        return self.identity.digest or self.identity.tag

    @cached_property
    def repository_url(self):
        # official repositories listed on docker hub have the 'library/' prefix
        if '/' in self.identity.repository:
            repository = self.identity.repository
        else:
            repository = f'library/{self.identity.repository}'

        return f'https://{self.registry_host}/v2/{repository}'

    @cached_property
    def session(self):
        session = self.auth_service.create_session(self.auth)
        session.headers['Accept'] = self.accept_header

        session.hooks = {
            'response': lambda r, **kwargs: r.raise_for_status()
        }

        return session

    @cached_property
    def manifest(self):
        url = f'{self.repository_url}/manifests/{self.reference}'
        manifest = self.session.get(url).json()

        if manifest.get('errors'):
            raise RuntimeError(manifest['errors'])

        return manifest

    def download_layer(self, layer, path):
        url = f'{self.repository_url}/blobs/{layer}'

        with self.session.get(url, stream=True) as response:
            with open(path, 'wb') as f:
                for chunk in response.iter_content(chunk_size=8 * 1024):
                    chunk and f.write(chunk)

    @property
    def latest_digest(self):
        url = f'{self.repository_url}/manifests/{self.reference}'
        return self.session.head(url).headers['Docker-Content-Digest']

    @property
    def layers(self):
        return tuple(l['digest'] for l in self.manifest['layers'])



from .base import RemoteImage
from .base import RemoteAuth

from .docker import DockerIdentity
from .docker import GCRAuth
from .docker import PublicDockerAuth
from .docker import RemoteDockerImage

__all__ = (
    'DockerIdentity',
    'GCRAuth',
    'PublicDockerAuth',
    'RemoteAuth',
    'RemoteDockerImage',
    'RemoteImage',
)



import os.path
import tarfile
import shutil


def extract_oci_layers(layers, destination):

    def is_whiteout(member):
        return os.path.basename(member.path).startswith('.wh.')

    for layer in layers:
        with tarfile.open(layer, errorlevel=0) as tar:
            # apply whiteouts to the destination
            members, whiteouts = separate(tar.getmembers(), is_whiteout)

            for whiteout in whiteouts:
                apply_whiteout(destination, whiteout.path)

            # extract files sans whiteouts
            tar.extractall(path=destination, members=members)


def separate(items, is_separate):
    a, b = [], []

    for item in items:
        if is_separate(item):
            b.append(item)
        else:
            a.append(item)

    return a, b


def apply_whiteout(destination, whiteout):
    if whiteout.endswith('.wh..wh..opq'):
        apply_opaque_whiteout(destination, whiteout)
    else:
        apply_simple_whiteout(destination, whiteout)


def apply_simple_whiteout(destination, whiteout):
    target = os.path.join(
        os.path.dirname(whiteout),
        os.path.basename(whiteout)[4:]
    )

    if not os.path.exists(target):
        return

    if os.path.isdir(target):
        shutil.rmtree(target)
    else:
        os.unlink(target)


def apply_opaque_whiteout(destination, whiteout):
    target = os.path.join(destination, os.path.dirname(whiteout))

    if not os.path.exists(target):
        return

    for root, dirs, files in os.walk(target):
        for d in dirs:
            shutil.rmtree(os.path.join(root, d))
        for f in files:
            os.unlink(os.path.join(root, f))



import click
import os
import ostree.local
import ostree.remote
import shutil
import sys
import tempfile


def default_cache_dir():
    if os.getuid() == 0:
        return '/var/cache/ostree'
    else:
        return os.path.expanduser('~/.cache/seantis/ostree')


def available_protocols():
    protocols = list(ostree.remote.RemoteImage.protocols.keys())
    protocols.sort()

    return protocols


def postmortem_hook(type, value, tb):
    """ Optionally used as excepthook, when post-mortem debugging is used. """

    if hasattr(sys, 'ps1') or not sys.stderr.isatty():
        # we are in interactive mode or we don't have a tty-like
        # device, so we call the default hook
        sys.__excepthook__(type, value, tb)
    else:
        import traceback
        # we are NOT in interactive mode, print the exception...
        traceback.print_exception(type, value, tb)
        print()

        import pdb
        pdb.post_mortem(tb)


@click.group()
@click.option(
    '--pdb/--no-pdb', 'postmortem', help="Enable post-mortem", default=False)
def cli(postmortem):

    if postmortem:
        sys.excepthook = postmortem_hook


def with_cache(fn):
    return click.option(
        '--cache',
        help="The cache directory. May be a path or the keyword 'no'",
        default=default_cache_dir()
    )(fn)


def with_auth(fn):
    return click.option(
        '--auth',
        help="A token or the path to a keyfile.",
        default=None
    )(fn)


def with_protocol(fn):
    return click.option(
        '--protocol', help="The image protocol in use (only docker for now)",
        default='docker', type=click.Choice(available_protocols())
    )(fn)


@cli.command()
@click.argument("name")
@click.argument("destination")
@click.option(
    '--force', help="Removes the destination if it exists",
    is_flag=True, default=False
)
@with_cache
@with_auth
@with_protocol
def pull(name, destination, force, cache, auth, protocol):
    """ Pulls an image from a registry and creates an OS tree.

    Example:

        ostree pull ubuntu:18.04 /var/machines/ubuntu

    The will download and extract the image with the given name into the
    destination. If the directory already exists the command will error.

    """

    if cache == 'no':
        cache = tempfile.TemporaryDirectory().name

    if force and os.path.isdir(destination):
        shutil.rmtree(destination)

    image = ostree.remote.RemoteImage.from_name(protocol, name, auth)
    local = ostree.local.Cache(cache)

    local.pull(image, destination)


@cli.command()
@click.argument("name")
@with_auth
@with_protocol
def digest(name, auth, protocol):
    """ Returns the latest digest of the given address.

    Note: If the digest is part of the address, then said digest is returned
    without any interaction with the remote repository.

    """
    image = ostree.remote.RemoteImage.from_name(protocol, name, auth)

    print(image.latest_digest)


@cli.command()
@with_cache
def purge(cache):
    """ Takes a cache directory and purges all layers no longer in use. """

    local = ostree.local.Cache(cache)
    local.purge()



import hashlib
import os

from itertools import chain
from ostree.oci import extract_oci_layers
from pathlib import Path


def ensure_folder(path):
    os.makedirs(path, exist_ok=True)
    return path


def link_path(parent_folder, destination):
    name = hashlib.md5(str(destination).encode('utf-8')).hexdigest()
    return parent_folder / name


def link_write(path, destination, layers):
    with path.open('w') as f:
        f.write(f'{destination}\n')
        f.writelines('\n'.join(layers))


def link_read(path):
    with path.open('r') as f:
        destination = Path(f.readline().strip())
        layers = f.read().strip().split('\n')

    return destination, layers


class LayersLink(object):

    def __init__(self, destination, layers):
        self.destination = destination
        self.layers = layers

    @classmethod
    def load(cls, path):
        with open(path, 'r') as f:
            return cls(
                destination=Path(f.readline().strip()),
                layers='\n'.split(f.read().strip())
            )


class Cache(object):

    def __init__(self, path):
        self.path = ensure_folder(Path(path))

    @property
    def layers(self):
        return ensure_folder(self.path / 'layers')

    @property
    def links(self):
        return ensure_folder(self.path / 'links')

    def pull_layer(self, image, layer):
        path = self.layers / layer

        if not path.exists():
            image.download_layer(layer, path)

        return path

    def pull(self, image, destination):
        destination = Path(destination)
        destination.mkdir(parents=True, exist_ok=False)

        # this seems like an obvious place to introduce threading, but note
        # that in real-world tests the difference didn't look like it was
        # worth it (even without cache, with cache there's hardly a difference)
        #
        # even if we did the docker thing and downloaded and extracted layers
        # in a pipeline we would probably not gain too much except for increase
        # in code complexity
        #
        # also note that the requests library session we use is not really
        # thread-safe, so each thread would also have to keep its own
        # connection to the server
        files = [self.pull_layer(image, layer) for layer in image.layers]
        extract_oci_layers(files, destination)

        # keep track of the layers in use
        link = link_path(self.links, destination)
        link_write(link, destination, image.layers)

    def purge(self):
        known = {}

        for path in self.links.iterdir():
            known.__setitem__(*link_read(path))

        for destination in tuple(known.keys()):
            if not destination.exists():
                link_path(self.links, destination).unlink()
                del known[destination]

        known_layers = set(chain(*known.values()))

        for layer in self.layers.iterdir():
            if layer.name not in known_layers:
                layer.unlink()



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

name = 'ostree'
description = (
    'Pulls containers and turns them into OS trees for systemd-nspawn.'
)
version = '0.1.0'


def get_long_description():
    readme = open('README.md').read()
    history = open('HISTORY.md').read()

    # cut the part before the description to avoid repetition on pypi
    readme = readme[readme.index(description) + len(description):]

    return '\n'.join((readme, history))


setup(
    name=name,
    version=version,
    description=description,
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url='http://github.com/seantis/ostree',
    author='Seantis GmbH',
    author_email='info@seantis.ch',
    license='MIT',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=name.split('.')[:-1],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'cached_property',
        'click',
        'google-auth',
        'requests',
    ],
    extras_require=dict(
        test=[
            'coverage',
            'flake8',
            'pytest',
        ],
    ),
    entry_points={
        'console_scripts': 'ostree=ostree.cli:cli'
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
    ]
)



import logging
import pytest
import shutil

from click.testing import CliRunner
from shlex import split
from ostree.cli import cli


def run(commandline):
    return CliRunner().invoke(cli, split(commandline), catch_exceptions=False)


@pytest.mark.parametrize('image, digest', [(
    'ubuntu:18.04',
    'sha256:b5309340de7a9a540cf6c0cba3eabdfb9c9bc5153026d37991fd0028180fc725'
), (
    'gcr.io/google-containers/etcd:3.2.24',
    'sha256:adf0009b3ea09d82fb15c7649a7436e0635cb51c74c5d58796db6c3c52713a1c'
)])
def test_digest(image, digest, caplog):
    # asking for the digest of an image we identiy *by* digest, we should
    # get the same digest back (ostree should still make a lookup though)
    caplog.set_level(logging.DEBUG, logger='urllib3')

    assert run(f'digest {image}@{digest}').output.strip() == digest
    assert 'Starting new HTTPS connection' in caplog.messages[0]


@pytest.mark.parametrize('image', (
    'ubuntu:18.04',
    'gcr.io/google-containers/etcd:3.2.24'
))
def test_pull(image, temppath):
    cache = temppath / 'cache'
    tree = temppath / 'tree'

    # pull the image
    result = run(f'pull {image} {tree} --cache={cache}')
    assert result.exit_code == 0

    # make sure there are downloaded layers
    assert sum(1 for p in (cache / 'layers').iterdir())

    # make sure the files were extracted
    assert sum(1 for p in tree.iterdir())
    assert (tree / 'bin').exists()

    # try to overwrite the image (this only works with --force)
    with pytest.raises(FileExistsError):
        result = run(f'pull {image} {tree} --cache={cache}')
        assert result.exit_code != 0

    result = run(f'pull {image} {tree} --cache={cache} --force')
    assert result.exit_code == 0


def test_purge(temppath):
    cache = temppath / 'cache'
    images = temppath / 'images'

    # pull two images with some shared layers
    run(f'pull alpine:3.8 {images}/alpine --cache={cache}')
    run(f'pull python:alpine3.8 {images}/python --cache={cache}')

    # we now have two linked files
    assert sum(1 for l in (cache / 'links').iterdir()) == 2

    # and a number of layers
    layers = sum(1 for l in (cache / 'layers').iterdir())
    assert layers

    # removing the base image should not result in any removed layers
    shutil.rmtree(images / 'alpine')
    run(f'purge --cache={cache}')

    assert sum(1 for l in (cache / 'links').iterdir()) == 1
    assert layers == sum(1 for l in (cache / 'layers').iterdir())

    # doing it the other way around howevers hould result in removed layers
    run(f'pull alpine:3.8 {images}/alpine --cache={cache}')
    shutil.rmtree(images / 'python')
    run(f'purge --cache={cache}')

    assert sum(1 for l in (cache / 'links').iterdir()) == 1
    assert layers > sum(1 for l in (cache / 'layers').iterdir())

    # removing the last image and purging the cache results in no layers
    shutil.rmtree(images / 'alpine')
    run(f'purge --cache={cache}')

    assert sum(1 for l in (cache / 'links').iterdir()) == 0
    assert sum(1 for l in (cache / 'layers').iterdir()) == 0



class RemoteImage(object):

    protocols = {}

    def __init_subclass__(cls, protocol, **kwargs):
        cls.protocols[protocol] = cls

        super().__init_subclass__(**kwargs)

    @classmethod
    def from_name(cls, protocol, name, auth=None):
        if protocol not in RemoteImage.protocols:
            raise NotImplementedError(f"Unknown protocol: {protocol}")

        return cls.protocols[protocol](name, auth=auth)


class RemoteAuth(object):

    hosts = {}

    def __init__(self, identity):
        self.identity = identity

    def __init_subclass__(cls, hosts, **kwargs):
        for host in hosts:
            cls.hosts[host] = cls

        super().__init_subclass__(**kwargs)

    @classmethod
    def from_identity(cls, identity):
        if identity.host not in cls.hosts:
            raise NotImplementedError(f"Unknown host: {identity.host}")

        return cls.hosts[identity.host](identity)
